#include "mbed.h"
#include "math.h"
#include "MPU9150.h"



/****************************************************************************
    MPU9150 Example Code
    Robotic wireless systems
    Evangelos Dermatas,
    1-4-2017

I2C Hardware setup:
 VDD ---------------------- 3.3V
 SDA ----------------------- PB_8
 SCL ----------------------- PB_9
 GND ---------------------- GND

*****************************************************************************

  NUCLEO-F334R8
  PINS
**WEELS SECTION

    LEFT WHEEL
        leftIN1         -> PC_10
        leftIN2         -> PC_11
        leftWenable     -> PC_12

    RIGHT WHEEL
        rightIN1        -> PA_13
        rightIN2        -> PA_14
        rightWenable    -> PA_15

**ULTASENSOR SECTION
// Distance estimation based on Ultasound sensor: 200ms

    TRIGGER
        ultraTrig       -> PC_6
    ECHO RECEIVER
        ultraEcho       -> PC_8

********************************************************************************************************************/

float sum = 0;
uint32_t sumCount = 0, mcount = 0;
char buffer[14];
MPU9150 MPU9150;
Timer t;
Serial pc(USBTX, USBRX); // tx, rx
// ********************************************************************************************************************

DigitalOut led(LED1);

//**************************************************
//**************  Ultrasound variables  ************
//**************************************************

DigitalOut      ultraTrig(PC_6) ;
InterruptIn     ultraEcho(PC_8) ;
Timer           EchoTime ;
Ticker          UltaTickerTrig ;
float           UltraDistance ;
float           UltraEchoTime ;


//**************************************************
//**************  Infrared variables  **************
//**************************************************

//AnalogIn      InfraPin(A0) ;
////**************************************************
//float InfraDistance(void)
//{
//    double volts = InfraPin.read() * 5;                // value from sensor * (5/1024)
//    return 65*pow(volts, -1.10);
//    // worked out from graph 65
//}




//**************************************************
//**************  Ultrasound functions  ************
//**************************************************

//**************************************************
void UltraTrigStart()
{
//    led=1 ;
    ultraTrig = 1 ;
    wait_us(12) ;
    ultraTrig = 0 ;
//    led=0 ;
}

//**************************************************
void UltraEchoStart()
{
    EchoTime.reset() ;
    EchoTime.start() ;
    led=1 ;
}


//**************************************************
void UltraEchoEnd()
{
    EchoTime.stop() ;
    UltraEchoTime = EchoTime.read();
    UltraDistance = UltraEchoTime * (340.0*50) ;   // estimation in centimeters
    led=0 ;
}




// ********************************************************************************************************************
int main()
{

    UltaTickerTrig.attach(&UltraTrigStart, 0.5 );   // every 200ms
    ultraEcho.rise(&UltraEchoStart) ;
    ultraEcho.fall(&UltraEchoEnd) ;



    pc.baud(9600);

    pc.printf( "Hello three wheel robotic car\r\n" ) ;
    pc.printf("CPU SystemCoreClock is %d Hz\r\n", SystemCoreClock);

    //Set up I2C in fast mode: 400 kHz
    i2c.frequency(400000);


    t.start();

    // Read the WHO_AM_I register, this is a good test of communication
    uint8_t whoami = MPU9150.readByte(MPU9150_ADDRESS, WHO_AM_I_MPU9150);  // Read WHO_AM_I register for MPU-9250
    pc.printf("I AM 0x%x\n\r", whoami);
    pc.printf("I SHOULD BE 0x68\n\r");

    if (whoami == 0x68) { // WHO_AM_I should be 0x68
        pc.printf("MPU9150 WHO_AM_I is 0x%x\n\r", whoami);
        pc.printf("MPU9150 is online...\n\r");
        wait(1);

        MPU9150.MPU9150SelfTest(SelfTest);
        pc.printf("x-axis self test: acceleration trim within %f % of factory value\n\r", SelfTest[0]);
        pc.printf("y-axis self test: acceleration trim within %f % of factory value\n\r", SelfTest[1]);
        pc.printf("z-axis self test: acceleration trim within %f % of factory value\n\r", SelfTest[2]);
        pc.printf("x-axis self test: gyration trim within %f % of factory value\n\r", SelfTest[3]);
        pc.printf("y-axis self test: gyration trim within %f % of factory value\n\r", SelfTest[4]);
        pc.printf("z-axis self test: gyration trim within %f % of factory value\n\r", SelfTest[5]);
        wait(1);
        MPU9150.resetMPU9150(); // Reset registers to default in preparation for device calibration
        MPU9150.calibrateMPU9150(gyroBias, accelBias); // Calibrate gyro and accelerometers, load biases in bias registers
        pc.printf("x gyro bias = %f\n\r", gyroBias[0]);
        pc.printf("y gyro bias = %f\n\r", gyroBias[1]);
        pc.printf("z gyro bias = %f\n\r", gyroBias[2]);
        pc.printf("x accel bias = %f\n\r", accelBias[0]);
        pc.printf("y accel bias = %f\n\r", accelBias[1]);
        pc.printf("z accel bias = %f\n\r", accelBias[2]);
        wait(1);
        MPU9150.initMPU9150();
        pc.printf("MPU9150 initialized for active data mode....\n\r"); // Initialize device for active mode read of acclerometer, gyroscope, and temperature
        MPU9150.initAK8975A(magCalibration);
        pc.printf("AK8975 initialized for active data mode....\n\r"); // Initialize device for active mode read of magnetometer
    } else {
        pc.printf("Could not connect to MPU9150: \n\r");
        pc.printf("%#x \n",  whoami);
        while(1) ; // Loop forever if communication doesn't happen
    }

    uint8_t MagRate = 10; // set magnetometer read rate in Hz; 10 to 100 (max) Hz are reasonable values
    MPU9150.getAres(); // Get accelerometer sensitivity
    MPU9150.getGres(); // Get gyro sensitivity
    mRes = 10.*1229./4096.; // Conversion from 1229 microTesla full scale (4096) to 12.29 Gauss full scale
    // So far, magnetometer bias is calculated and subtracted here manually, should construct an algorithm to do it automatically
    // like the gyro and accelerometer biases
    magbias[0] = -5.;   // User environmental x-axis correction in milliGauss
    magbias[1] = -95.;  // User environmental y-axis correction in milliGauss
    magbias[2] = -260.; // User environmental z-axis correction in milliGauss


    while(1) {

        // If intPin goes high, all data registers have new data
        if(MPU9150.readByte(MPU9150_ADDRESS, INT_STATUS) & 0x01) {  // On interrupt, check if data ready interrupt

            MPU9150.readAccelData(accelCount);  // Read the x/y/z adc values
            // Now we'll calculate the accleration value into actual g's
            ax = (float)accelCount[0]*aRes; // - accelBias[0];  // get actual g value, this depends on scale being set
            ay = (float)accelCount[1]*aRes; // - accelBias[1];
            az = (float)accelCount[2]*aRes; // - accelBias[2];

            MPU9150.readGyroData(gyroCount);  // Read the x/y/z adc values
            // Calculate the gyro value into actual degrees per second
            gx = (float)gyroCount[0]*gRes; // - gyroBias[0];  // get actual gyro value, this depends on scale being set
            gy = (float)gyroCount[1]*gRes; // - gyroBias[1];
            gz = (float)gyroCount[2]*gRes; // - gyroBias[2];

            mcount++;
            if (mcount > 200/MagRate) {  // this is a poor man's way of setting the magnetometer read rate (see below)
                MPU9150.readMagData(magCount);  // Read the x/y/z adc values
                // Calculate the magnetometer values in milliGauss
                // Include factory calibration per data sheet and user environmental corrections
                mx = (float)magCount[0]*mRes*magCalibration[0] - magbias[0];  // get actual magnetometer value, this depends on scale being set
                my = (float)magCount[1]*mRes*magCalibration[1] - magbias[1];
                mz = (float)magCount[2]*mRes*magCalibration[2] - magbias[2];
                mcount = 0;
            }
        }

        Now = t.read_us();
        deltat = (float)((Now - lastUpdate)/1000000.0f) ; // set integration time by time elapsed since last filter update
        lastUpdate = Now;

        sum += deltat;
        sumCount++;

        // Pass gyro rate as rad/s
//  MPU9150.MadgwickQuaternionUpdate(ax, ay, az, gx*PI/180.0f, gy*PI/180.0f, gz*PI/180.0f,  my,  mx, mz);
        MPU9150.MahonyQuaternionUpdate(ax, ay, az, gx*PI/180.0f, gy*PI/180.0f, gz*PI/180.0f, my, mx, mz);

        // Serial print and/or display at 0.5 s rate independent of data rates
        delt_t = t.read_ms() - count;
        if (delt_t > 500) { // update LCD once per half-second independent of read rate

            tempCount = MPU9150.readTempData();  // Read the adc values
            temperature = ((float) tempCount) / 340.0f + 36.53f; // Temperature in degrees Centigrade

            pc.printf( "A=%7.1f %7.1f %7.1f | G=%5.1f %5.1f %5.1f |  T=%3.1f | D=%6.2f %7.4f | \r\n",
                       1000*ax,1000*ay,1000*az, gx, gy, gz,temperature,UltraDistance, UltraEchoTime ) ;

            count = t.read_ms();

            if(count > 1<<21) {
                t.start(); // start the timer over again if ~30 minutes has passed
                count = 0;
                deltat= 0;
                lastUpdate = t.read_us();
            }
            sum = 0;
            sumCount = 0;
        }
    }
}
